#include <string>
#include <iostream>
#include <fstream>
#include <cstring>
#include "bmp.h"
using namespace std;

// Write a single short integer to a file in PC format (low byte comes first)
bool write_PC_short(ofstream &f, short num)
{

    char byte;
    // Write the low bytes first, shift 8 more bytes each time
    for (int i = 0; i <= 1; i++) {
        byte = num&0xFF;
        if (!f.put(byte)) {
            cout << "write_bitmap_file: error writing int to file" << endl;
            return false;
        }
        num >>= 8;
    }
}

// Write a single integer to a file in PC format (lowest to highest bytes)
bool write_PC_int(ofstream &f, int num)
{
    char byte;
    // Write the low bytes first, shift 8 more bytes each time
    for (int i = 0; i <= 3; i++) {
        byte = num&0xFF;
        if (!f.put(byte)) {
            cout << "write_bitmap_file: error writing int to file" << endl;
            return false;
        }
        num >>=8;
    }
}

// Each pixel gets written 3 times - one each for RGB.  This simulates gray
bool write_three_copies(ofstream &bmpfile, unsigned char pixel)
{
    int k;
    // write 3 copies of each pixel, one for each color
    for (int k=0; k < 3; k++) {
        if (!bmpfile.put(pixel)) {
            cout << "Write bitmap file: error writing data" << endl;
            return false;
        }
    }
}

// writes a grayscale .bmp file.  Pass it a file name, width and height
// in pixels, and the image data.  Image data is assumed to be one byte
// per pixel, with grayscale data from 0-255.  0 is black, 255 is white
bool write_bitmap_file(string bmpfn, int w, int h, unsigned char *data)
{
    ofstream bmpfile(bmpfn.c_str(), ios::out |ios::binary); // output file
    bmfh_struct  bmpfh;   // file header
    bmih_struct bmih;     // info header
    int roundwidth((w+3)/4*4);   // width and height must be multiples of 4
    int roundheight((h+3)/4*4);

    // check that file opened ok
    if (!bmpfile.is_open()) {
        cout << "Write bitmap file: error opening file " << bmpfn << endl;
        return false;
    }

    // set all values in the file header
    strcpy(bmpfh.bfType, BMF_DEFAULT_TYPE);
    // file contains file header, info header, and 3 bytes per pixel:
    bmpfh.bfSize = BMF_DEFAULT_SIZE+BMIH_DEFAULT_SIZE+roundwidth*roundheight*3;
    bmpfh.bfReserved1 = 0;
    bmpfh.bfReserved2 = 0; 
    bmpfh.bfOffBits = BMF_DEFAULT_SIZE+BMIH_DEFAULT_SIZE;

    bmpfile.put(bmpfh.bfType[0]);
    bmpfile.put(bmpfh.bfType[1]);
    if (!write_PC_int(bmpfile,bmpfh.bfSize)) return false;
    if (!write_PC_short(bmpfile,bmpfh.bfReserved1)) return false;
    if (!write_PC_short(bmpfile,bmpfh.bfReserved2)) return false;
    if (!write_PC_int(bmpfile,bmpfh.bfOffBits)) return false;
    
    // set values in information header
    bmih.biSize = BMIH_DEFAULT_SIZE;   // info header size - should be 40
    bmih.biWidth = roundwidth;     // Width of image in pixels
    bmih.biHeight = roundheight;   // Height of image in pixels
    bmih.biPlanes = 1;             // default - 0
    bmih.biBitCount= 24;           // 3 bytes - one each for RGB
    bmih.biCompression = 0;        // Compression type - 0 for our use
    bmih.biSizeImage=0;            // set to 0 if no compression
    bmih.biXPelsPerMeter=0;        // default to 0
    bmih.biYPelsPerMeter=0;        // default to 0
    bmih.biClrUsed=0;              // set to 0 to calculate from BitCount
    bmih.biClrImportant=0;         // 0 for all colors important

    // write information header to file
    if (!write_PC_int(bmpfile, bmih.biSize)) return false;
    if (!write_PC_int(bmpfile, bmih.biWidth)) return false;
    if (!write_PC_int(bmpfile, bmih.biHeight)) return false;
    if (!write_PC_short(bmpfile, bmih.biPlanes)) return false;
    if (!write_PC_short(bmpfile, bmih.biBitCount)) return false;
    if (!write_PC_int(bmpfile, bmih.biCompression)) return false;
    if (!write_PC_int(bmpfile, bmih.biSizeImage)) return false;
    if (!write_PC_int(bmpfile, bmih.biXPelsPerMeter)) return false;
    if (!write_PC_int(bmpfile, bmih.biYPelsPerMeter)) return false;
    if (!write_PC_int(bmpfile, bmih.biClrUsed)) return false;
    if (!write_PC_int(bmpfile, bmih.biClrImportant)) return false;

    int i, j;
    // Write image data - start with padding if not a multiple of 4
    for (i =  roundheight-1; i >= h; i--)  {
        for (j=0; j < roundwidth; j++) {
            if (!write_three_copies(bmpfile, 0)) return false;
        }
    }
    // Write actual data.  Each byte in each row must be written 3 times
    for (i = h-1; i >= 0; i--)  {
        for (j =  0; j < w; j++) {
            write_three_copies(bmpfile, data[i*w+j]);
        }        
        // padding if width is not a multiple of 4
        for (j=w; j <roundwidth; j++) {
            write_three_copies(bmpfile, 0);
        }
    }

    bmpfile.close();
    return true;
}


